@@ -2,36 +2,6 @@
 static Loop *default_loop = NULL;
 
 
-static PyTypeObject LoopCountersResultType;
-
-static PyStructSequence_Field loop_counters_result_fields[] = {
-    {"eio_init", ""},
-    {"req_init", ""},
-    {"handle_init", ""},
-    {"stream_init", ""},
-    {"tcp_init", ""},
-    {"udp_init", ""},
-    {"pipe_init", ""},
-    {"tty_init", ""},
-    {"poll_init", ""},
-    {"prepare_init", ""},
-    {"check_init", ""},
-    {"idle_init", ""},
-    {"async_init", ""},
-    {"timer_init", ""},
-    {"process_init", ""},
-    {"fs_event_init", ""},
-    {NULL}
-};
-
-static PyStructSequence_Desc loop_counters_result_desc = {
-    "loop_counters_result",
-    NULL,
-    loop_counters_result_fields,
-    16
-};
-
-
 static void
 _loop_cleanup(void)
 {
@@ -39,6 +9,34 @@ _loop_cleanup(void)
 }
 
 
+static int
+init_loop(Loop *loop, int is_default)
+{
+    uv_loop_t *uv_loop;
+
+    uv_loop = NULL;
+
+    if (is_default) {
+        uv_loop = uv_default_loop();
+    } else {
+        uv_loop = uv_loop_new();
+    }
+
+    if (!uv_loop) {
+        PyErr_NoMemory();
+        return -1;
+    }
+    uv_loop->data = (void *)loop;
+
+    loop->uv_loop = uv_loop;
+    loop->is_default = is_default;
+    loop->weakreflist = NULL;
+    loop->excepthook_cb = NULL;
+
+    return 0;
+}
+
+
 static PyObject *
 new_loop(PyTypeObject *type, PyObject *args, PyObject *kwargs, int is_default)
 {
@@ -53,10 +51,9 @@ new_loop(PyTypeObject *type, PyObject *args, PyObject *kwargs, int is_default)
             if (!default_loop) {
                 return NULL;
             }
-            default_loop->uv_loop = uv_default_loop();
-            default_loop->uv_loop->data = (void *)default_loop;
-            default_loop->is_default = 1;
-            default_loop->weakreflist = NULL;
+            if (init_loop(default_loop, True) != 0) {
+                return NULL;
+            }
             Py_AtExit(_loop_cleanup);
         }
         Py_INCREF(default_loop);
@@ -66,47 +63,55 @@ new_loop(PyTypeObject *type, PyObject *args, PyObject *kwargs, int is_default)
         if (!self) {
             return NULL;
         }
-        self->uv_loop = uv_loop_new();
-        self->uv_loop->data = (void *)self;
-        self->is_default = 0;
-        self->weakreflist = NULL;
+        if (init_loop(self, False) != 0) {
+            return NULL;
+        }
         return (PyObject *)self;
     }
 }
 
 
 static PyObject *
-Loop_func_run(Loop *self)
+Loop_func_run(Loop *self, PyObject *args)
 {
-    Py_BEGIN_ALLOW_THREADS
-    uv_run(self->uv_loop);
-    Py_END_ALLOW_THREADS
-    if (PyErr_Occurred()) {
-        PyErr_WriteUnraisable(Py_None);
+    int r;
+    int mode;
+
+    mode = UV_RUN_DEFAULT;
+
+    if (!PyArg_ParseTuple(args, "|i:run", &mode)) {
+        return NULL;
     }
-    Py_RETURN_NONE;
+
+    if (mode != UV_RUN_DEFAULT && mode != UV_RUN_ONCE && mode != UV_RUN_NOWAIT) {
+        PyErr_SetString(PyExc_ValueError, "invalid mode specified");
+        return NULL;
+    }
+
+    Py_BEGIN_ALLOW_THREADS
+    r = uv_run(self->uv_loop, mode);
+    Py_END_ALLOW_THREADS
+
+    if (PyErr_Occurred()) {
+        handle_uncaught_exception(self);
+    }
+
+    return PyBool_FromLong((long)r);
 }
 
 
 static PyObject *
-Loop_func_run_once(Loop *self)
+Loop_func_stop(Loop *self)
 {
-    int r;
-
-    Py_BEGIN_ALLOW_THREADS
-    r = uv_run_once(self->uv_loop);
-    Py_END_ALLOW_THREADS
-    if (PyErr_Occurred()) {
-        PyErr_WriteUnraisable(Py_None);
-    }
-    return PyBool_FromLong((long)r);
+    uv_stop(self->uv_loop);
+    Py_RETURN_NONE;
 }
 
 
 static PyObject *
 Loop_func_now(Loop *self)
 {
-    return PyLong_FromDouble(uv_now(self->uv_loop));
+    return PyLong_FromUnsignedLongLong(uv_now(self->uv_loop));
 }
 
 
@@ -121,14 +126,14 @@ Loop_func_update_time(Loop *self)
 static void
 walk_cb(uv_handle_t* handle, void* arg)
 {
+    PyObject *result;
     PyObject *callback = (PyObject *)arg;
-    PyObject *result, *obj;
-    if (handle->data != NULL) {
-        obj = (PyObject *)handle->data;
+    PyObject *obj = (PyObject *)handle->data;
+    if (obj && Py_REFCNT(obj) > 0) {
         Py_INCREF(obj);
         result = PyObject_CallFunctionObjArgs(callback, obj, NULL);
         if (result == NULL) {
-            PyErr_WriteUnraisable(callback);
+            handle_uncaught_exception(((Handle *)obj)->loop);
         }
         Py_DECREF(obj);
         Py_XDECREF(result);
@@ -158,18 +163,153 @@ Loop_func_walk(Loop *self, PyObject *args)
 
 
 static PyObject *
-Loop_func_default_loop(PyObject *cls)
+Loop_func_fileno(Loop *self)
 {
-    UNUSED_ARG(cls);
-    return new_loop(&LoopType, NULL, NULL, 1);
+    return PyInt_FromLong((long)uv_backend_fd(self->uv_loop));
 }
 
 
 static PyObject *
-Loop_active_handles_get(Loop *self, void *closure)
+Loop_func_get_timeout(Loop *self)
 {
-    UNUSED_ARG(closure);
-    return PyInt_FromLong((long)self->uv_loop->active_handles);
+    return PyFloat_FromDouble(uv_backend_timeout(self->uv_loop)/1000.0);
+}
+
+
+static void
+threadpool_work_cb(uv_work_t *req)
+{
+    PyGILState_STATE gstate = PyGILState_Ensure();
+    WorkRequest *pyreq;
+    PyObject *result;
+
+    ASSERT(req);
+    pyreq = (WorkRequest *)req->data;
+
+    result = PyObject_CallFunctionObjArgs(pyreq->work_cb, NULL);
+    if (result == NULL) {
+        ASSERT(PyErr_Occurred());
+        PyErr_Print();
+    }
+    Py_XDECREF(result);
+
+    PyGILState_Release(gstate);
+}
+
+static void
+threadpool_done_cb(uv_work_t *req, int status)
+{
+    PyGILState_STATE gstate = PyGILState_Ensure();
+    uv_err_t err;
+    WorkRequest *pyreq;
+    Loop *loop;
+    PyObject *result, *errorno;
+
+    ASSERT(req);
+    loop = (Loop *)req->loop->data;
+    pyreq = (WorkRequest *)req->data;
+
+    if (pyreq->done_cb) {
+        if (status < 0) {
+            err = uv_last_error(req->loop);
+            errorno = PyInt_FromLong((long)err.code);
+        } else {
+            errorno = Py_None;
+            Py_INCREF(Py_None);
+        }
+
+        result = PyObject_CallFunctionObjArgs(pyreq->done_cb, errorno, NULL);
+        if (result == NULL) {
+            handle_uncaught_exception(loop);
+        }
+        Py_XDECREF(result);
+        Py_DECREF(errorno);
+    }
+
+    Py_DECREF(pyreq->work_cb);
+    Py_XDECREF(pyreq->done_cb);
+    pyreq->work_cb = NULL;
+    pyreq->done_cb = NULL;
+    ((Request *)pyreq)->req = NULL;
+    Py_DECREF(pyreq);
+    PyMem_Free(req);
+
+    /* Refcount was increased in queue_work */
+    Py_DECREF(loop);
+
+    PyGILState_Release(gstate);
+}
+
+static PyObject *
+Loop_func_queue_work(Loop *self, PyObject *args)
+{
+    int r;
+    uv_work_t *req;
+    WorkRequest *pyreq;
+    PyObject *work_cb, *done_cb;
+
+    req = NULL;
+    pyreq = NULL;
+    done_cb = NULL;
+
+    if (!PyArg_ParseTuple(args, "O|O:queue_work", &work_cb, &done_cb)) {
+        return NULL;
+    }
+
+    if (!PyCallable_Check(work_cb)) {
+        PyErr_SetString(PyExc_TypeError, "a callable is required");
+        return NULL;
+    }
+
+    if (done_cb != NULL && !PyCallable_Check(done_cb)) {
+        PyErr_SetString(PyExc_TypeError, "done_cb must be a callable");
+        return NULL;
+    }
+
+    req = PyMem_Malloc(sizeof *req);
+    if (!req) {
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    pyreq = (WorkRequest *)PyObject_CallObject((PyObject *)&WorkRequestType, NULL);
+    if (!pyreq) {
+        PyErr_NoMemory();
+        goto error;
+    }
+    ((Request *)pyreq)->req = (uv_req_t *)req;
+    pyreq->work_cb = work_cb;
+    pyreq->done_cb = done_cb;
+
+    Py_INCREF(work_cb);
+    Py_XINCREF(done_cb);
+    req->data = (void *)pyreq;
+
+    r = uv_queue_work(self->uv_loop, req, threadpool_work_cb, threadpool_done_cb);
+    if (r != 0) {
+        RAISE_UV_EXCEPTION(self->uv_loop, PyExc_Exception);
+        goto error;
+    }
+
+    Py_INCREF(self);
+
+    Py_INCREF(pyreq);
+    return (PyObject *)pyreq;
+
+error:
+    Py_XDECREF(pyreq);
+    Py_DECREF(work_cb);
+    Py_XDECREF(done_cb);
+    PyMem_Free(req);
+    return NULL;
+}
+
+
+static PyObject *
+Loop_func_default_loop(PyObject *cls)
+{
+    UNUSED_ARG(cls);
+    return new_loop(&LoopType, NULL, NULL, 1);
 }
 
 
@@ -185,58 +325,6 @@ Loop_default_get(Loop *self, void *closure)
 }
 
 
-static PyObject *
-get_counter_value(uint64_t value)
-{
-    PyObject *val;
-#if SIZEOF_TIME_T > SIZEOF_LONG
-    val = PyLong_FromLongLong((PY_LONG_LONG)value);
-#else
-    val = PyInt_FromLong((long)value);
-#endif
-    return val;
-}
-
-static PyObject *
-Loop_counters_get(Loop *self, void *closure)
-{
-    int pos;
-    uv_counters_t* uv_counters;
-    PyObject *counters;
-
-    UNUSED_ARG(closure);
-    uv_counters = &self->uv_loop->counters;
-
-    counters =  PyStructSequence_New(&LoopCountersResultType);
-    if (!counters) {
-        PyErr_NoMemory();
-        return NULL;
-    }
-
-    pos = 0;
-
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->eio_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->req_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->handle_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->stream_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->tcp_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->udp_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->pipe_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->tty_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->poll_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->prepare_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->check_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->idle_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->async_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->timer_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->process_init));
-    PyStructSequence_SET_ITEM(counters, pos++, get_counter_value(uv_counters->fs_event_init));
-
-    return counters;
-
-}
-
-
 static PyObject *
 Loop_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
@@ -264,6 +352,7 @@ static void
 Loop_tp_dealloc(Loop *self)
 {
     if (self->uv_loop) {
+        self->uv_loop->data = NULL;
         uv_loop_delete(self->uv_loop);
     }
     if (self->weakreflist != NULL) {
@@ -313,23 +402,56 @@ Loop_dict_set(Loop *self, PyObject* val, void* c)
 }
 
 
+static PyObject*
+Loop_excepthook_get(Loop *self, void* c)
+{
+    UNUSED_ARG(c);
+    return self->excepthook_cb;
+}
+
+
+static int
+Loop_excepthook_set(Loop *self, PyObject* val, void* c)
+{
+    PyObject* tmp;
+
+    UNUSED_ARG(c);
+
+    if (val == NULL) {
+        PyErr_SetString(PyExc_TypeError, "excepthook may not be deleted");
+        return -1;
+    }
+    if (val != Py_None && !PyCallable_Check(val)) {
+        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
+        return -1;
+    }
+    tmp = self->excepthook_cb;
+    Py_INCREF(val);
+    self->excepthook_cb = val;
+    Py_XDECREF(tmp);
+    return 0;
+}
+
+
 static PyMethodDef
 Loop_tp_methods[] = {
-    { "run", (PyCFunction)Loop_func_run, METH_NOARGS, "Run the event loop." },
-    { "run_once", (PyCFunction)Loop_func_run_once, METH_NOARGS, "Run a single event loop iteration, waiting for events if necessary." },
+    { "run", (PyCFunction)Loop_func_run, METH_VARARGS, "Run the event loop." },
+    { "stop", (PyCFunction)Loop_func_stop, METH_NOARGS, "Stop running the event loop." },
     { "now", (PyCFunction)Loop_func_now, METH_NOARGS, "Return event loop time, expressed in nanoseconds." },
     { "update_time", (PyCFunction)Loop_func_update_time, METH_NOARGS, "Update event loop's notion of time by querying the kernel." },
     { "walk", (PyCFunction)Loop_func_walk, METH_VARARGS, "Walk all handles in the loop." },
+    { "fileno", (PyCFunction)Loop_func_fileno, METH_NOARGS, "Get the loop backend file descriptor." },
+    { "get_timeout", (PyCFunction)Loop_func_get_timeout, METH_NOARGS, "Get the poll timeout, or -1 for no timeout." },
     { "default_loop", (PyCFunction)Loop_func_default_loop, METH_CLASS|METH_NOARGS, "Instantiate the default loop." },
+    { "queue_work", (PyCFunction)Loop_func_queue_work, METH_VARARGS, "Queue the given function to be run in the thread pool." },
     { NULL }
 };
 
 
 static PyGetSetDef Loop_tp_getsets[] = {
     {"__dict__", (getter)Loop_dict_get, (setter)Loop_dict_set, NULL},
-    {"active_handles", (getter)Loop_active_handles_get, NULL, "Number of active handles in this loop", NULL},
     {"default", (getter)Loop_default_get, NULL, "Is this the default loop?", NULL},
-    {"counters", (getter)Loop_counters_get, NULL, "Loop counters", NULL},
+    {"excepthook", (getter)Loop_excepthook_get, (setter)Loop_excepthook_set, "Loop uncaught exception handler", NULL},
     {NULL}
 };
 
@@ -354,7 +476,7 @@ static PyTypeObject LoopType = {
     0,                                                              /*tp_getattro*/
     0,                                                              /*tp_setattro*/
     0,                                                              /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,  /*tp_flags*/
     0,                                                              /*tp_doc*/
     (traverseproc)Loop_tp_traverse,                                 /*tp_traverse*/
     (inquiry)Loop_tp_clear,                                         /*tp_clear*/
