@@ -1,7 +1,4 @@
 
-static PyObject* PyExc_PipeError;
-
-
 static void
 on_pipe_connection(uv_stream_t* server, int status)
 {
@@ -25,7 +22,7 @@ on_pipe_connection(uv_stream_t* server, int status)
 
     result = PyObject_CallFunctionObjArgs(self->on_new_connection_cb, self, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->on_new_connection_cb);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(py_errorno);
@@ -39,18 +36,14 @@ static void
 on_pipe_client_connection(uv_connect_t *req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    stream_req_data_t* req_data;
     Pipe *self;
     PyObject *callback, *result, *py_errorno;
     ASSERT(req);
 
-    req_data = (stream_req_data_t *)req->data;
     self = (Pipe *)req->handle->data;
-    callback = req_data->callback;
+    callback = (PyObject *)req->data;
 
     ASSERT(self);
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
 
     if (status != 0) {
         uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
@@ -62,16 +55,17 @@ on_pipe_client_connection(uv_connect_t *req, int status)
 
     result = PyObject_CallFunctionObjArgs(callback, self, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(callback);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(py_errorno);
 
     Py_DECREF(callback);
-    PyMem_Free(req_data);
     PyMem_Free(req);
 
+    /* Refcount was increased in the caller function */
     Py_DECREF(self);
+
     PyGILState_Release(gstate);
 }
 
@@ -96,7 +90,7 @@ on_pipe_read2(uv_pipe_t* handle, int nread, uv_buf_t buf, uv_handle_type pending
         data = PyBytes_FromStringAndSize(buf.base, nread);
         py_errorno = Py_None;
         Py_INCREF(Py_None);
-    } else if (nread < 0) {
+    } else {
         data = Py_None;
         Py_INCREF(Py_None);
         err = uv_last_error(UV_HANDLE_LOOP(self));
@@ -105,18 +99,13 @@ on_pipe_read2(uv_pipe_t* handle, int nread, uv_buf_t buf, uv_handle_type pending
 
     result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, data, py_pending, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->on_read_cb);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(data);
     Py_DECREF(py_pending);
     Py_DECREF(py_errorno);
 
-    /* In case of error libuv may not call alloc_cb */
-    if (buf.base != NULL) {
-        PyMem_Free(buf.base);
-    }
-
     Py_DECREF(self);
     PyGILState_Release(gstate);
 }
@@ -128,6 +117,7 @@ Pipe_func_bind(Pipe *self, PyObject *args)
     int r;
     char *name;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "s:bind", &name)) {
@@ -153,6 +143,7 @@ Pipe_func_listen(Pipe *self, PyObject *args)
     backlog = 128;
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O|i:listen", &callback, &backlog)) {
@@ -190,14 +181,22 @@ Pipe_func_accept(Pipe *self, PyObject *args)
     int r;
     PyObject *client;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:accept", &client)) {
         return NULL;
     }
 
-    if (!PyObject_IsSubclass((PyObject *)client->ob_type, (PyObject *)&StreamType)) {
-        PyErr_SetString(PyExc_TypeError, "Only stream objects are supported for accept");
+    if (PyObject_IsSubclass((PyObject *)client->ob_type, (PyObject *)&StreamType)) {
+        if (UV_HANDLE(client)->type != UV_TCP && UV_HANDLE(client)->type != UV_NAMED_PIPE) {
+            PyErr_SetString(PyExc_TypeError, "Only TCP and Pipe objects are supported for accept");
+            return NULL;
+        }
+    } else if (PyObject_IsSubclass((PyObject *)client->ob_type, (PyObject *)&UDPType)) {
+        /* empty */
+    } else {
+        PyErr_SetString(PyExc_TypeError, "Only Stream and UDP objects are supported for accept");
         return NULL;
     }
 
@@ -216,9 +215,9 @@ Pipe_func_connect(Pipe *self, PyObject *args)
 {
     char *name;
     uv_connect_t *connect_req = NULL;
-    stream_req_data_t *req_data = NULL;
     PyObject *callback;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "sO:connect", &name, &callback)) {
@@ -230,34 +229,28 @@ Pipe_func_connect(Pipe *self, PyObject *args)
         return NULL;
     }
 
-    connect_req = (uv_connect_t *)PyMem_Malloc(sizeof(uv_connect_t));
+    Py_INCREF(callback);
+
+    connect_req = PyMem_Malloc(sizeof *connect_req);
     if (!connect_req) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
-    req_data->callback = callback;
-    connect_req->data = (void *)req_data;
+    connect_req->data = (void *)callback;
 
     uv_pipe_connect(connect_req, (uv_pipe_t *)UV_HANDLE(self), name, on_pipe_client_connection);
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
 error:
+    Py_DECREF(callback);
     if (connect_req) {
         PyMem_Free(connect_req);
     }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
     return NULL;
 }
 
@@ -267,6 +260,7 @@ Pipe_func_open(Pipe *self, PyObject *args)
 {
     int fd;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:open", &fd)) {
@@ -285,6 +279,7 @@ Pipe_func_pending_instances(Pipe *self, PyObject *args)
     /* This function applies to Windows only */
     int count;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:pending_instances", &count)) {
@@ -305,6 +300,7 @@ Pipe_func_start_read2(Pipe *self, PyObject *args)
 
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start_read2", &callback)) {
@@ -334,97 +330,54 @@ Pipe_func_start_read2(Pipe *self, PyObject *args)
 static PyObject *
 Pipe_func_write2(Pipe *self, PyObject *args)
 {
-    int i, r, buf_count;
-    Py_buffer pbuf;
+    uv_buf_t buf;
+    Py_buffer *view;
     PyObject *callback, *send_handle;
-    uv_buf_t *bufs = NULL;
-    uv_write_t *wr = NULL;
-    stream_req_data_t *req_data = NULL;
-    stream_write_data_t *write_data = NULL;
 
-    buf_count = 0;
     callback = Py_None;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "s*O|O:write2", &pbuf, &send_handle, &callback)) {
+    view = PyMem_Malloc(sizeof *view);
+    if (!view) {
+        PyErr_NoMemory();
         return NULL;
     }
 
-    if (!PyObject_TypeCheck(send_handle, &TCPType) && !PyObject_TypeCheck(send_handle, &PipeType)) {
-        PyErr_SetString(PyExc_TypeError, "Only TCP and Pipe objects are supported for write2");
+#ifdef PYUV_PYTHON3
+    if (!PyArg_ParseTuple(args, "y*O|O:write", view, &send_handle, &callback)) {
+#else
+    if (!PyArg_ParseTuple(args, "s*O|O:write", view, &send_handle, &callback)) {
+#endif
         return NULL;
     }
 
+    if (PyObject_IsSubclass((PyObject *)send_handle->ob_type, (PyObject *)&StreamType)) {
+        if (UV_HANDLE(send_handle)->type != UV_TCP && UV_HANDLE(send_handle)->type != UV_NAMED_PIPE) {
+            PyErr_SetString(PyExc_TypeError, "Only TCP and Pipe objects are supported for write2");
+            goto error;
+        }
+    } else if (PyObject_IsSubclass((PyObject *)send_handle->ob_type, (PyObject *)&UDPType)) {
+        /* empty */
+    } else {
+        PyErr_SetString(PyExc_TypeError, "Only Stream and UDP objects are supported");
+        goto error;
+    }
+
     if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyBuffer_Release(&pbuf);
         PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    wr = (uv_write_t *)PyMem_Malloc(sizeof(uv_write_t));
-    if (!wr) {
-        PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
+    buf = uv_buf_init(view->buf, view->len);
 
-    write_data = (stream_write_data_t *) PyMem_Malloc(sizeof(stream_write_data_t));
-    if (!write_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
-    req_data->callback = callback;
-    wr->data = (void *)req_data;
-
-    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t));
-    if (!bufs) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    bufs[0] = uv_buf_init(pbuf.buf, pbuf.len);
-    buf_count = 1;
-
-    write_data->bufs = bufs;
-    write_data->buf_count = buf_count;
-    write_data->view = pbuf;
-    req_data->data = write_data;
-
-    r = uv_write2(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, (uv_stream_t *)UV_HANDLE(send_handle), on_stream_write);
-    if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
-        goto error;
-    }
-
-    Py_RETURN_NONE;
+    return pyuv_stream_write((Stream *)self, view, &buf, 1, callback, send_handle);
 
 error:
-    PyBuffer_Release(&pbuf);
-    if (bufs) {
-        for (i = 0; i < buf_count; i++) {
-            PyMem_Free(bufs[i].base);
-        }
-        PyMem_Free(bufs);
-    }
-    if (write_data) {
-        PyMem_Free(write_data);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (wr) {
-        PyMem_Free(wr);
-    }
-    return NULL;
+        PyBuffer_Release(view);
+        PyMem_Free(view);
+        return NULL;
 }
 
 
@@ -432,42 +385,24 @@ static int
 Pipe_tp_init(Pipe *self, PyObject *args, PyObject *kwargs)
 {
     int r;
-    uv_pipe_t *uv_pipe;
     Loop *loop;
-    PyObject *tmp = NULL;
     PyObject *ipc = Py_False;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_PipeError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!|O!:__init__", &LoopType, &loop, &PyBool_Type, &ipc)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
-
-    uv_pipe = PyMem_Malloc(sizeof(uv_pipe_t));
-    if (!uv_pipe) {
-        PyErr_NoMemory();
-        Py_DECREF(loop);
-        return -1;
-    }
-
-    r = uv_pipe_init(UV_HANDLE_LOOP(self), uv_pipe, (ipc == Py_True) ? 1 : 0);
+    r = uv_pipe_init(loop->uv_loop, (uv_pipe_t *)UV_HANDLE(self), (ipc == Py_True) ? 1 : 0);
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
-        Py_DECREF(loop);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_PipeError);
         return -1;
     }
-    uv_pipe->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_pipe;
+
+    initialize_handle(HANDLE(self), loop);
 
     return 0;
 }
@@ -476,10 +411,24 @@ Pipe_tp_init(Pipe *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Pipe_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Pipe *self = (Pipe *)StreamType.tp_new(type, args, kwargs);
-    if (!self) {
+    Pipe *self;
+    uv_pipe_t *uv_pipe;
+
+    uv_pipe = PyMem_Malloc(sizeof *uv_pipe);
+    if (!uv_pipe) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (Pipe *)StreamType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_pipe);
+        return NULL;
+    }
+
+    uv_pipe->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_pipe;
+
     return (PyObject *)self;
 }
 
@@ -536,7 +485,7 @@ static PyTypeObject PipeType = {
     0,                                                             /*tp_getattro*/
     0,                                                             /*tp_setattro*/
     0,                                                             /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                       /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                                                             /*tp_doc*/
     (traverseproc)Pipe_tp_traverse,                                /*tp_traverse*/
     (inquiry)Pipe_tp_clear,                                        /*tp_clear*/
