@@ -1,7 +1,4 @@
 
-static PyObject* PyExc_PollError;
-
-
 static void
 on_poll_callback(uv_poll_t *handle, int status, int events)
 {
@@ -30,7 +27,7 @@ on_poll_callback(uv_poll_t *handle, int status, int events)
 
     result = PyObject_CallFunctionObjArgs(self->callback, self, py_events, py_errorno,NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->callback);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
 
@@ -47,6 +44,7 @@ Poll_func_start(Poll *self, PyObject *args)
 
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "iO:start", &events, &callback)) {
@@ -78,6 +76,7 @@ Poll_func_stop(Poll *self)
 {
     int r;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_poll_stop((uv_poll_t *)UV_HANDLE(self));
@@ -94,21 +93,13 @@ Poll_func_stop(Poll *self)
 
 
 static PyObject *
-Poll_slow_get(Poll *self, void *closure)
+Poll_func_fileno(Poll *self)
 {
-    UNUSED_ARG(closure);
-    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
-#ifdef PYUV_WINDOWS
-    #define UV_HANDLE_POLL_SLOW  0x02000000 /* copied from src/win/internal.h */
-    if (!(UV_HANDLE(self)->flags & UV_HANDLE_POLL_SLOW)) {
-        Py_RETURN_FALSE;
-    } else {
-        Py_RETURN_TRUE;
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+    if (uv_is_closing(UV_HANDLE(self))) {
+        return PyInt_FromLong(-1);
     }
-    #undef UV_HANDLE_POLL_SLOW
-#else
-    Py_RETURN_FALSE;
-#endif
+    return PyInt_FromLong(self->fd);
 }
 
 
@@ -117,43 +108,25 @@ Poll_tp_init(Poll *self, PyObject *args, PyObject *kwargs)
 {
     int r;
     long fd;
-    uv_poll_t *uv_poll = NULL;
     Loop *loop;
-    PyObject *tmp;
-
-    tmp = NULL;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_PollError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!l:__init__", &LoopType, &loop, &fd)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
-
-    uv_poll = PyMem_Malloc(sizeof(uv_poll_t));
-    if (!uv_poll) {
-        PyErr_NoMemory();
-        Py_DECREF(loop);
-        return -1;
-    }
-
-    r = uv_poll_init_socket(UV_HANDLE_LOOP(self), uv_poll, (uv_os_sock_t)fd);
+    r = uv_poll_init_socket(loop->uv_loop, (uv_poll_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PollError);
-        Py_DECREF(loop);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_PollError);
         return -1;
     }
-    uv_poll->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_poll;
+
+    self->fd = fd;
+
+    initialize_handle(HANDLE(self), loop);
 
     return 0;
 }
@@ -162,10 +135,24 @@ Poll_tp_init(Poll *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Poll_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Poll *self = (Poll *)HandleType.tp_new(type, args, kwargs);
-    if (!self) {
+    Poll *self;
+    uv_poll_t *uv_poll;
+
+    uv_poll = PyMem_Malloc(sizeof *uv_poll);
+    if (!uv_poll) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (Poll *)HandleType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_poll);
+        return NULL;
+    }
+
+    uv_poll->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_poll;
+
     return (PyObject *)self;
 }
 
@@ -192,16 +179,11 @@ static PyMethodDef
 Poll_tp_methods[] = {
     { "start", (PyCFunction)Poll_func_start, METH_VARARGS, "Start the Poll." },
     { "stop", (PyCFunction)Poll_func_stop, METH_NOARGS, "Stop the Poll." },
+    { "fileno", (PyCFunction)Poll_func_fileno, METH_NOARGS, "File descriptor being monitored." },
     { NULL }
 };
 
 
-static PyGetSetDef Poll_tp_getsets[] = {
-    {"slow", (getter)Poll_slow_get, 0, "Indicates if the handle is running in slow mode (Windows)", NULL},
-    {NULL}
-};
-
-
 static PyTypeObject PollType = {
     PyVarObject_HEAD_INIT(NULL, 0)
     "pyuv.Poll",                                                    /*tp_name*/
@@ -222,7 +204,7 @@ static PyTypeObject PollType = {
     0,                                                              /*tp_getattro*/
     0,                                                              /*tp_setattro*/
     0,                                                              /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,  /*tp_flags*/
     0,                                                              /*tp_doc*/
     (traverseproc)Poll_tp_traverse,                                 /*tp_traverse*/
     (inquiry)Poll_tp_clear,                                         /*tp_clear*/
@@ -232,7 +214,7 @@ static PyTypeObject PollType = {
     0,                                                              /*tp_iternext*/
     Poll_tp_methods,                                                /*tp_methods*/
     0,                                                              /*tp_members*/
-    Poll_tp_getsets,                                                /*tp_getsets*/
+    0,                                                              /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
     0,                                                              /*tp_descr_get*/
@@ -243,4 +225,3 @@ static PyTypeObject PollType = {
     Poll_tp_new,                                                    /*tp_new*/
 };
 
-
