@@ -1,13 +1,4 @@
 
-/*
- * NOTE: Since libuv uses pipes to communicate processes and pipes
- * are buffered, things will not go so well. Replacing the comunication
- * mechanism with a pty could do the job, perhaps. References:
- *
- * http://stackoverflow.com/questions/4057985/disabling-stdout-buffering-of-a-forked-process
- * http://stackoverflow.com/questions/2055918/forcing-a-program-to-flush-its-standard-output-when-redirected
- */
-
 /* Container for standard IO */
 static int
 StdIO_tp_init(StdIO *self, PyObject *args, PyObject *kwargs)
@@ -67,7 +58,6 @@ StdIO_stream_get(StdIO *self, void *closure)
 {
     UNUSED_ARG(closure);
     if (!self->stream) {
-        Py_INCREF(Py_None);
         Py_RETURN_NONE;
     } else {
         return self->stream;
@@ -87,7 +77,7 @@ static PyObject *
 StdIO_flags_get(StdIO *self, void *closure)
 {
     UNUSED_ARG(closure);
-    return PyInt_FromLong((long)self->fd);
+    return PyInt_FromLong((long)self->flags);
 }
 
 
@@ -178,8 +168,21 @@ static PyTypeObject StdIOType = {
 
 /* Process handle */
 
-static PyObject* PyExc_ProcessError;
+#define RAISE_IF_SPAWNED(obj, retval)                                               \
+    do {                                                                            \
+        if ((obj)->spawned) {                                                       \
+            PyErr_SetString(PyExc_ProcessError, "Process already spawned");         \
+            return retval;                                                          \
+        }                                                                           \
+    } while(0)                                                                      \
 
+#define RAISE_IF_NOT_SPAWNED(obj, retval)                                           \
+    do {                                                                            \
+        if (!((obj)->spawned)) {                                                    \
+            PyErr_SetString(PyExc_ProcessError, "Process was not spawned");         \
+            return retval;                                                          \
+        }                                                                           \
+    } while(0)                                                                      \
 
 static void
 on_process_exit(uv_process_t *process, int exit_status, int term_signal)
@@ -192,22 +195,20 @@ on_process_exit(uv_process_t *process, int exit_status, int term_signal)
     self = (Process *)process->data;
     ASSERT(self);
 
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
-
     py_exit_status = PyInt_FromLong(exit_status);
     py_term_signal = PyInt_FromLong(term_signal);
 
     if (self->on_exit_cb != Py_None) {
         result = PyObject_CallFunctionObjArgs(self->on_exit_cb, self, py_exit_status, py_term_signal, NULL);
         if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_exit_cb);
+            handle_uncaught_exception(HANDLE(self)->loop);
         }
         Py_XDECREF(result);
         Py_DECREF(py_exit_status);
         Py_DECREF(py_term_signal);
     }
 
+    /* Refcount was increased in the spawn function */
     Py_DECREF(self);
 
     PyGILState_Release(gstate);
@@ -223,7 +224,6 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     char **ptr, **process_args, **process_env;
     Py_ssize_t i, n, pos;
     PyObject *key, *value, *item, *tmp, *callback, *arguments, *env, *stdio, *ret;
-    uv_process_t *uv_process;
     uv_process_options_t options;
     uv_stdio_container_t *stdio_container;
 
@@ -235,10 +235,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     stdio_container = NULL;
     flags = uid = gid = stdio_count = 0;
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_ProcessError, "Process already spawned");
-        return NULL;
-    }
+    RAISE_IF_SPAWNED(self, NULL);
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|OOO!sIIiO:__init__", kwlist, &file, &callback, &arguments, &PyDict_Type, &env, &cwd, &uid, &gid, &flags, &stdio)) {
         return NULL;
@@ -266,7 +263,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     options.flags = flags;
     options.exit_cb = on_process_exit;
 
-    file2 = (char *) PyMem_Malloc(strlen(file) + 1);
+    file2 = PyMem_Malloc(strlen(file) + 1);
     if (!file2) {
         PyErr_NoMemory();
         ret = NULL;
@@ -277,7 +274,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
 
     if (arguments) {
         n = PySequence_Length(arguments);
-        process_args = (char **)PyMem_Malloc(sizeof(char *) * (n + 2));
+        process_args = PyMem_Malloc(sizeof *process_args * (n + 2));
         if (!process_args) {
             PyErr_NoMemory();
             PyMem_Free(file2);
@@ -293,7 +290,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
                 ret = NULL;
                 goto cleanup;
             }
-            tmp_str = (char *) PyMem_Malloc(strlen(arg_str) + 1);
+            tmp_str = PyMem_Malloc(strlen(arg_str) + 1);
             if (!tmp_str) {
                 Py_DECREF(item);
                 ret = NULL;
@@ -306,7 +303,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
         }
         process_args[i+1] = NULL;
     } else {
-        process_args = (char **)PyMem_Malloc(sizeof(char *) * 2);
+        process_args = PyMem_Malloc(sizeof *process_args * 2);
         if (!process_args) {
             PyErr_NoMemory();
             PyMem_Free(file2);
@@ -319,7 +316,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     options.args = process_args;
 
     if (cwd) {
-        cwd2 = (char *) PyMem_Malloc(strlen(cwd) + 1);
+        cwd2 = PyMem_Malloc(strlen(cwd) + 1);
         if (!cwd2) {
             PyErr_NoMemory();
             ret = NULL;
@@ -332,7 +329,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     if (env) {
         n = PyDict_Size(env);
         if (n > 0) {
-            process_env = (char **)PyMem_Malloc(sizeof(char *) * (n + 1));
+            process_env = PyMem_Malloc(sizeof *process_env * (n + 1));
             if (!process_env) {
                 PyErr_NoMemory();
                 ret = NULL;
@@ -346,7 +343,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
                     goto cleanup;
                 }
                 len = strlen(key_str) + strlen(value_str) + 2;
-                tmp_str = (char *) PyMem_Malloc(len);
+                tmp_str = PyMem_Malloc(len);
                 if (!tmp_str) {
                     PyErr_NoMemory();
                     ret = NULL;
@@ -363,7 +360,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
 
     if (stdio) {
         n = PySequence_Length(stdio);
-        stdio_container = (uv_stdio_container_t *)PyMem_Malloc(sizeof(uv_stdio_container_t) * n);
+        stdio_container = PyMem_Malloc(sizeof *stdio_container * n);
         if (!stdio_container) {
             PyErr_NoMemory();
             ret = NULL;
@@ -391,20 +388,9 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     options.stdio = stdio_container;
     options.stdio_count = stdio_count;
 
-    uv_process = PyMem_Malloc(sizeof(uv_process_t));
-    if (!uv_process) {
-        PyErr_NoMemory();
-        ret = NULL;
-        goto cleanup;
-    }
-    uv_process->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_process;
-
-    r = uv_spawn(UV_HANDLE_LOOP(self), uv_process, options);
+    r = uv_spawn(UV_HANDLE_LOOP(self), (uv_process_t *)UV_HANDLE(self), options);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_ProcessError);
-        PyMem_Free(uv_process);
-        UV_HANDLE(self) = NULL;
         ret = NULL;
         goto cleanup;
     }
@@ -419,28 +405,32 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     self->stdio = stdio;
     Py_XDECREF(tmp);
 
+    HANDLE(self)->initialized = True;
+    self->spawned = True;
+
     ret = Py_None;
 
+    /* Increase refcount so that object is not removed before the exit callback is called */
+    Py_INCREF(self);
+
 cleanup:
     if (options.args) {
         for (ptr = options.args; *ptr != NULL; ptr++) {
             PyMem_Free(*ptr);
         }
-        PyMem_Free(options.args);
-    }
-    if (options.cwd) {
-        PyMem_Free(options.cwd);
     }
+
     if (options.env) {
         for (ptr = options.env; *ptr != NULL; ptr++) {
             PyMem_Free(*ptr);
         }
-        PyMem_Free(options.env);
-    }
-    if (options.stdio) {
-        PyMem_Free(options.stdio);
     }
 
+    PyMem_Free(options.args);
+    PyMem_Free(options.cwd);
+    PyMem_Free(options.env);
+    PyMem_Free(options.stdio);
+
     Py_XINCREF(ret);
     return ret;
 }
@@ -451,10 +441,9 @@ Process_func_kill(Process *self, PyObject *args)
 {
     int signum, r;
 
-    if (UV_HANDLE_CLOSED(self)) {
-        PyErr_SetString(PyExc_ProcessError, "Process wasn't spawned yet");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+    RAISE_IF_NOT_SPAWNED(self, NULL);
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:kill", &signum)) {
         return NULL;
@@ -475,7 +464,7 @@ Process_pid_get(Process *self, void *closure)
 {
     UNUSED_ARG(closure);
 
-    if (UV_HANDLE_CLOSED(self)) {
+    if (!HANDLE(self)->initialized || !self->spawned) {
         Py_RETURN_NONE;
     }
     return PyInt_FromLong((long)((uv_process_t *)UV_HANDLE(self))->pid);
@@ -485,24 +474,20 @@ Process_pid_get(Process *self, void *closure)
 static int
 Process_tp_init(Process *self, PyObject *args, PyObject *kwargs)
 {
-    PyObject *tmp = NULL;
     Loop *loop;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_ProcessError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!:__init__", &LoopType, &loop)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
+    initialize_handle(HANDLE(self), loop);
+
+    /* uv_process_t handles are not initialized explicitly, so workaround it. See tp_dealloc for the rest */
+    HANDLE(self)->initialized = False;
 
     return 0;
 }
@@ -520,14 +505,37 @@ Process_func_disable_stdio_inheritance(PyObject *cls)
 static PyObject *
 Process_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Process *self = (Process *)HandleType.tp_new(type, args, kwargs);
-    if (!self) {
+    Process *self;
+    uv_process_t *uv_process;
+
+    uv_process = PyMem_Malloc(sizeof *uv_process);
+    if (!uv_process) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (Process *)HandleType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_process);
+        return NULL;
+    }
+
+    uv_process->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_process;
+    self->spawned = False;
+
     return (PyObject *)self;
 }
 
 
+static void
+Process_tp_dealloc(Process *self)
+{
+    HANDLE(self)->initialized = self->spawned;
+    HandleType.tp_dealloc((PyObject *)self);
+}
+
+
 static int
 Process_tp_traverse(Process *self, visitproc visit, void *arg)
 {
@@ -568,7 +576,7 @@ static PyTypeObject ProcessType = {
     "pyuv.Process",                                                 /*tp_name*/
     sizeof(Process),                                                /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    0,                                                              /*tp_dealloc*/
+    (destructor)Process_tp_dealloc,                                 /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -583,7 +591,7 @@ static PyTypeObject ProcessType = {
     0,                                                              /*tp_getattro*/
     0,                                                              /*tp_setattro*/
     0,                                                              /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,  /*tp_flags*/
     0,                                                              /*tp_doc*/
     (traverseproc)Process_tp_traverse,                              /*tp_traverse*/
     (inquiry)Process_tp_clear,                                      /*tp_clear*/
