@@ -1,46 +1,60 @@
 
-/*
- * This is not a true signal handler but it does the job. It uses a Prepare handle, which
- * gets executed one per loop iteration, before blocking for I/O, to call PyErr_CheckSignals.
- * That will cause standard registered signal handlers to be called by Python.
- */
-
-static PyObject* PyExc_SignalError;
-
-
 static void
-on_signal_callback(uv_prepare_t *handle, int status)
+on_signal_callback(uv_signal_t *handle, int signum)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     Signal *self;
+    PyObject *result;
 
     ASSERT(handle);
-    ASSERT(status == 0);
 
     self = (Signal *)handle->data;
     ASSERT(self);
+    /* Object could go out of scope in the callback, increase refcount to avoid it */
+    Py_INCREF(self);
 
-    if (PyErr_CheckSignals() < 0 && PyErr_Occurred()) {
-        PyErr_Print();
+    result = PyObject_CallFunctionObjArgs(self->callback, self, PyInt_FromLong((long)signum), NULL);
+    if (result == NULL) {
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
+    Py_XDECREF(result);
 
+    Py_DECREF(self);
     PyGILState_Release(gstate);
 }
 
 
 static PyObject *
-Signal_func_start(Signal *self)
+Signal_func_start(Signal *self, PyObject *args)
 {
-    int r;
+    int r, signum;
+    PyObject *tmp, *callback;
 
+    tmp = NULL;
+
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_prepare_start((uv_prepare_t *)UV_HANDLE(self), on_signal_callback);
+    if (!PyArg_ParseTuple(args, "Oi:start", &callback, &signum)) {
+        return NULL;
+    }
+
+    if (!PyCallable_Check(callback)) {
+        PyErr_SetString(PyExc_TypeError, "a callable is required");
+        return NULL;
+    }
+
+    r = uv_signal_start((uv_signal_t *)UV_HANDLE(self), (uv_signal_cb)on_signal_callback, signum);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_SignalError);
         return NULL;
     }
 
+    tmp = self->callback;
+    Py_INCREF(callback);
+    self->callback = callback;
+    Py_XDECREF(tmp);
+
     Py_RETURN_NONE;
 }
 
@@ -50,9 +64,10 @@ Signal_func_stop(Signal *self)
 {
     int r;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_prepare_stop((uv_prepare_t *)UV_HANDLE(self));
+    r = uv_signal_stop((uv_signal_t *)UV_HANDLE(self));
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_SignalError);
         return NULL;
@@ -66,41 +81,23 @@ static int
 Signal_tp_init(Signal *self, PyObject *args, PyObject *kwargs)
 {
     int r;
-    uv_prepare_t *uv_prepare = NULL;
     Loop *loop;
-    PyObject *tmp = NULL;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_SignalError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!:__init__", &LoopType, &loop)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
-
-    uv_prepare = PyMem_Malloc(sizeof(uv_prepare_t));
-    if (!uv_prepare) {
-        PyErr_NoMemory();
-        Py_DECREF(loop);
-        return -1;
-    }
-
-    r = uv_prepare_init(UV_HANDLE_LOOP(self), uv_prepare);
+    r = uv_signal_init(loop->uv_loop, (uv_signal_t *)UV_HANDLE(self));
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_SignalError);
-        Py_DECREF(loop);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_SignalError);
         return -1;
     }
-    uv_prepare->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_prepare;
+
+    initialize_handle(HANDLE(self), loop);
 
     return 0;
 }
@@ -109,10 +106,24 @@ Signal_tp_init(Signal *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Signal_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Signal *self = (Signal *)HandleType.tp_new(type, args, kwargs);
-    if (!self) {
+    Signal *self;
+    uv_signal_t *uv_signal;
+
+    uv_signal = PyMem_Malloc(sizeof *uv_signal);
+    if (!uv_signal) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (Signal *)HandleType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_signal);
+        return NULL;
+    }
+
+    uv_signal->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_signal;
+
     return (PyObject *)self;
 }
 
@@ -120,6 +131,7 @@ Signal_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 Signal_tp_traverse(Signal *self, visitproc visit, void *arg)
 {
+    Py_VISIT(self->callback);
     HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
@@ -128,6 +140,7 @@ Signal_tp_traverse(Signal *self, visitproc visit, void *arg)
 static int
 Signal_tp_clear(Signal *self)
 {
+    Py_CLEAR(self->callback);
     HandleType.tp_clear((PyObject *)self);
     return 0;
 }
@@ -135,8 +148,8 @@ Signal_tp_clear(Signal *self)
 
 static PyMethodDef
 Signal_tp_methods[] = {
-    { "start", (PyCFunction)Signal_func_start, METH_NOARGS, "Start the Signal." },
-    { "stop", (PyCFunction)Signal_func_stop, METH_NOARGS, "Stop the Signal." },
+    { "start", (PyCFunction)Signal_func_start, METH_VARARGS, "Start the Signal handle." },
+    { "stop", (PyCFunction)Signal_func_stop, METH_NOARGS, "Stop the Signal handle." },
     { NULL }
 };
 
@@ -161,7 +174,7 @@ static PyTypeObject SignalType = {
     0,                                                              /*tp_getattro*/
     0,                                                              /*tp_setattro*/
     0,                                                              /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,  /*tp_flags*/
     0,                                                              /*tp_doc*/
     (traverseproc)Signal_tp_traverse,                               /*tp_traverse*/
     (inquiry)Signal_tp_clear,                                       /*tp_clear*/
@@ -182,4 +195,3 @@ static PyTypeObject SignalType = {
     Signal_tp_new,                                                  /*tp_new*/
 };
 
-
