@@ -1,27 +1,18 @@
 
-static PyObject* PyExc_StreamError;
-
-
-typedef struct {
-    uv_buf_t *bufs;
-    int buf_count;
-    Py_buffer view;
-} stream_write_data_t;
-
 typedef struct {
     PyObject *callback;
-    stream_write_data_t *data;
-} stream_req_data_t;
+    PyObject *send_handle;
+    Py_buffer *views;
+    int view_count;
+} stream_write_data_t;
 
 
 static uv_buf_t
 on_stream_alloc(uv_stream_t* handle, size_t suggested_size)
 {
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    uv_buf_t buf = uv_buf_init(PyMem_Malloc(suggested_size), suggested_size);
+    static char slab[PYUV_SLAB_SIZE];
     UNUSED_ARG(handle);
-    PyGILState_Release(gstate);
-    return buf;
+    return uv_buf_init(slab, sizeof(slab));
 }
 
 
@@ -29,18 +20,14 @@ static void
 on_stream_shutdown(uv_shutdown_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    stream_req_data_t* req_data;
     uv_err_t err;
     Stream *self;
     PyObject *callback, *result, *py_errorno;
 
-    req_data = (stream_req_data_t *)req->data;
     self = (Stream *)req->handle->data;
-    callback = req_data->callback;
+    callback = (PyObject *)req->data;
 
     ASSERT(self);
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
 
     if (callback != Py_None) {
         if (status < 0) {
@@ -52,17 +39,18 @@ on_stream_shutdown(uv_shutdown_t* req, int status)
         }
         result = PyObject_CallFunctionObjArgs(callback, self, py_errorno, NULL);
         if (result == NULL) {
-            PyErr_WriteUnraisable(callback);
+            handle_uncaught_exception(HANDLE(self)->loop);
         }
         Py_XDECREF(result);
         Py_DECREF(py_errorno);
     }
 
     Py_DECREF(callback);
-    PyMem_Free(req_data);
     PyMem_Free(req);
 
+    /* Refcount was increased in the caller function */
     Py_DECREF(self);
+
     PyGILState_Release(gstate);
 }
 
@@ -85,26 +73,23 @@ on_stream_read(uv_stream_t* handle, int nread, uv_buf_t buf)
         data = PyBytes_FromStringAndSize(buf.base, nread);
         py_errorno = Py_None;
         Py_INCREF(Py_None);
-    } else if (nread < 0) {
+    } else {
         data = Py_None;
         Py_INCREF(Py_None);
         err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong((long)err.code);
+        /* Stop reading, otherwise an assert blows up on unix */
+        uv_read_stop(handle);
     }
 
     result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, data, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->on_read_cb);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(data);
     Py_DECREF(py_errorno);
 
-    /* In case of error libuv may not call alloc_cb */
-    if (buf.base != NULL) {
-        PyMem_Free(buf.base);
-    }
-
     Py_DECREF(self);
     PyGILState_Release(gstate);
 }
@@ -115,22 +100,19 @@ on_stream_write(uv_write_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     int i;
-    stream_req_data_t* req_data;
-    stream_write_data_t* write_data;
+    stream_write_data_t* req_data;
     Stream *self;
-    PyObject *callback, *result, *py_errorno;
+    PyObject *callback, *send_handle, *result, *py_errorno;
     uv_err_t err;
 
     ASSERT(req);
 
-    req_data = (stream_req_data_t *)req->data;
-    write_data = req_data->data;
+    req_data = (stream_write_data_t *)req->data;
     self = (Stream *)req->handle->data;
     callback = req_data->callback;
+    send_handle = req_data->send_handle;
 
     ASSERT(self);
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
 
     if (callback != Py_None) {
         if (status < 0) {
@@ -142,26 +124,24 @@ on_stream_write(uv_write_t* req, int status)
         }
         result = PyObject_CallFunctionObjArgs(callback, self, py_errorno, NULL);
         if (result == NULL) {
-            PyErr_WriteUnraisable(callback);
+            handle_uncaught_exception(HANDLE(self)->loop);
         }
         Py_XDECREF(result);
         Py_DECREF(py_errorno);
     }
 
-    if (write_data->buf_count == 1) {
-        PyBuffer_Release(&write_data->view);
-    } else {
-        for (i = 0; i < write_data->buf_count; i++) {
-            PyMem_Free(write_data->bufs[i].base);
-        }
-    }
-    PyMem_Free(write_data->bufs);
-    PyMem_Free(write_data);
     Py_DECREF(callback);
+    Py_XDECREF(send_handle);
+    for (i = 0; i < req_data->view_count; i++) {
+        PyBuffer_Release(&req_data->views[i]);
+    }
+    PyMem_Free(req_data->views);
     PyMem_Free(req_data);
     PyMem_Free(req);
 
+    /* Refcount was increased in the caller function */
     Py_DECREF(self);
+
     PyGILState_Release(gstate);
 }
 
@@ -171,30 +151,24 @@ Stream_func_shutdown(Stream *self, PyObject *args)
 {
     int r;
     uv_shutdown_t *req = NULL;
-    stream_req_data_t *req_data = NULL;
     PyObject *callback = Py_None;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "|O:shutdown", &callback)) {
         return NULL;
     }
 
-    req = (uv_shutdown_t*) PyMem_Malloc(sizeof(uv_shutdown_t));
+    Py_INCREF(callback);
+
+    req = PyMem_Malloc(sizeof *req);
     if (!req) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
-    req_data->callback = callback;
-    req->data = (void *)req_data;
+    req->data = (void *)callback;
 
     r = uv_shutdown(req, (uv_stream_t *)UV_HANDLE(self), on_stream_shutdown);
     if (r != 0) {
@@ -202,16 +176,14 @@ Stream_func_shutdown(Stream *self, PyObject *args)
         goto error;
     }
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
 error:
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (req) {
-        PyMem_Free(req);
-    }
+    Py_DECREF(callback);
+    PyMem_Free(req);
     return NULL;
 }
 
@@ -224,6 +196,7 @@ Stream_func_start_read(Stream *self, PyObject *args)
 
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start_read", &callback)) {
@@ -255,6 +228,7 @@ Stream_func_stop_read(Stream *self)
 {
     int r;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_read_stop((uv_stream_t *)UV_HANDLE(self));
@@ -270,133 +244,112 @@ Stream_func_stop_read(Stream *self)
 }
 
 
-static PyObject *
-Stream_func_write(Stream *self, PyObject *args)
+static INLINE PyObject *
+pyuv_stream_write(Stream *self, Py_buffer *views, uv_buf_t *bufs, int buf_count, PyObject *callback, PyObject *send_handle)
 {
-    int r, buf_count;
-    Py_buffer pbuf;
-    PyObject *callback;
-    uv_buf_t *bufs = NULL;
+    int i, r;
     uv_write_t *wr = NULL;
-    stream_req_data_t *req_data = NULL;
-    stream_write_data_t *write_data = NULL;
+    stream_write_data_t *req_data = NULL;
 
-    buf_count = 0;
-    callback = Py_None;
+    Py_INCREF(callback);
+    Py_XINCREF(send_handle);
 
-    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
-
-    if (!PyArg_ParseTuple(args, "s*|O:write", &pbuf, &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyBuffer_Release(&pbuf);
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    wr = (uv_write_t *)PyMem_Malloc(sizeof(uv_write_t));
+    wr = PyMem_Malloc(sizeof *wr);
     if (!wr) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
+    req_data = PyMem_Malloc(sizeof *req_data);
     if (!req_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    write_data = (stream_write_data_t *) PyMem_Malloc(sizeof(stream_write_data_t));
-    if (!write_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
     req_data->callback = callback;
+    req_data->send_handle = send_handle;
+    req_data->views = views;
+    req_data->view_count = buf_count;
+
     wr->data = (void *)req_data;
 
-    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t));
-    if (!bufs) {
-        PyErr_NoMemory();
-        goto error;
+    if (send_handle) {
+        r = uv_write2(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, (uv_stream_t *)UV_HANDLE(send_handle), on_stream_write);
+    } else {
+        r = uv_write(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, on_stream_write);
     }
 
-    bufs[0] = uv_buf_init(pbuf.buf, pbuf.len);
-    buf_count = 1;
-
-    write_data->bufs = bufs;
-    write_data->buf_count = buf_count;
-    write_data->view = pbuf;
-    req_data->data = write_data;
-
-    r = uv_write(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, on_stream_write);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_StreamError);
         goto error;
     }
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
 error:
-    PyBuffer_Release(&pbuf);
-    if (bufs) {
-        PyMem_Free(bufs);
-    }
-    if (write_data) {
-        PyMem_Free(write_data);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (wr) {
-        PyMem_Free(wr);
+    Py_DECREF(callback);
+    Py_XDECREF(send_handle);
+    for (i = 0; i < buf_count; i++) {
+        PyBuffer_Release(&views[i]);
     }
+    PyMem_Free(views);
+    PyMem_Free(req_data);
+    PyMem_Free(wr);
     return NULL;
 }
 
 
-static Py_ssize_t
-iter_guess_size(PyObject *o, Py_ssize_t defaultvalue)
+static PyObject *
+Stream_func_write(Stream *self, PyObject *args)
 {
-    PyObject *ro;
-    Py_ssize_t rv;
+    uv_buf_t buf;
+    Py_buffer *view;
+    PyObject *callback = Py_None;
 
-    /* try o.__length_hint__() */
-    ro = PyObject_CallMethod(o, "__length_hint__", NULL);
-    if (ro == NULL) {
-        /* whatever the error is, clear it and return the default */
-        PyErr_Clear();
-        return defaultvalue;
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    view = PyMem_Malloc(sizeof *view);
+    if (!view) {
+        PyErr_NoMemory();
+        return NULL;
     }
-    rv = PyLong_Check(ro) ? PyLong_AsSsize_t(ro) : defaultvalue;
-    Py_DECREF(ro);
-    return rv;
+
+#ifdef PYUV_PYTHON3
+    if (!PyArg_ParseTuple(args, "y*|O:write", view, &callback)) {
+#else
+    if (!PyArg_ParseTuple(args, "s*|O:write", view, &callback)) {
+#endif
+        return NULL;
+    }
+
+    if (callback != Py_None && !PyCallable_Check(callback)) {
+        PyBuffer_Release(view);
+        PyMem_Free(view);
+        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
+        return NULL;
+    }
+
+    buf = uv_buf_init(view->buf, view->len);
+
+    return pyuv_stream_write(self, view, &buf, 1, callback, NULL);
 }
 
+
 static PyObject *
 Stream_func_writelines(Stream *self, PyObject *args)
 {
-    int i, r, buf_count;
-    char *data_str, *tmp;
-    const char *default_encoding;
-    Py_buffer pbuf;
-    Py_ssize_t data_len, n;
-    PyObject *callback, *seq, *iter, *item, *encoded;
-    uv_buf_t tmpbuf;
-    uv_buf_t *bufs, *new_bufs;
-    uv_write_t *wr = NULL;
-    stream_req_data_t *req_data = NULL;
-    stream_write_data_t *write_data = NULL;
+    int r, buf_count;
+    Py_buffer *views;
+    PyObject *callback, *seq, *ret;
+    uv_buf_t *bufs;
 
-    buf_count = 0;
-    bufs = new_bufs = NULL;
     callback = Py_None;
-    default_encoding = PyUnicode_GetDefaultEncoding();
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O|O:writelines", &seq, &callback)) {
@@ -408,153 +361,18 @@ Stream_func_writelines(Stream *self, PyObject *args)
         return NULL;
     }
 
-    wr = (uv_write_t *)PyMem_Malloc(sizeof(uv_write_t));
-    if (!wr) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    write_data = (stream_write_data_t *) PyMem_Malloc(sizeof(stream_write_data_t));
-    if (!write_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
-    req_data->callback = callback;
-    wr->data = (void *)req_data;
-
-    iter = PyObject_GetIter(seq);
-    if (iter == NULL) {
-        goto error;
-    }
-
-    n = iter_guess_size(iter, 8);   /* if we can't get the size hint, preallocate 8 slots */
-    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t) * n);
-    if (!bufs) {
-        Py_DECREF(iter);
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    i = 0;
-    while (1) {
-        item = PyIter_Next(iter);
-        if (item == NULL) {
-            if (PyErr_Occurred()) {
-                Py_DECREF(iter);
-                goto error;
-            } else {
-                /* StopIteration */
-                break;
-            }
-        }
-
-        if (PyUnicode_Check(item)) {
-            encoded = PyUnicode_AsEncodedString(item, default_encoding, "strict");
-            if (encoded == NULL) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                goto error;
-            }
-            data_str = PyBytes_AS_STRING(encoded);
-            data_len = PyBytes_GET_SIZE(encoded);
-            tmp = (char *) PyMem_Malloc(data_len);
-            if (!tmp) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyErr_NoMemory();
-                goto error;
-            }
-            memcpy(tmp, data_str, data_len);
-            tmpbuf = uv_buf_init(tmp, data_len);
-        } else {
-            if (PyObject_GetBuffer(item, &pbuf, PyBUF_CONTIG_RO) < 0) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                goto error;
-            }
-            data_str = pbuf.buf;
-            data_len = pbuf.len;
-            tmp = (char *) PyMem_Malloc(data_len);
-            if (!tmp) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyBuffer_Release(&pbuf);
-                PyErr_NoMemory();
-                goto error;
-            }
-            memcpy(tmp, data_str, data_len);
-            tmpbuf = uv_buf_init(tmp, data_len);
-            PyBuffer_Release(&pbuf);
-        }
-
-        /* Check if we allocated enough space */
-        if (buf_count+1 < n) {
-            /* we have enough size */
-        } else {
-            /* preallocate 8 more slots */
-            n += 8;
-            new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * n);
-            if (!new_bufs) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyErr_NoMemory();
-                goto error;
-            }
-            bufs = new_bufs;
-        }
-        bufs[i] = tmpbuf;
-        i++;
-        buf_count++;
-        Py_DECREF(item);
-    }
-    Py_DECREF(iter);
-
-    /* we may have over allocated space, shrink it to the minimum required */
-    new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * buf_count);
-    if (!new_bufs) {
-        PyErr_NoMemory();
-        goto error;
-    }
-    bufs = new_bufs;
-
-    write_data->bufs = bufs;
-    write_data->buf_count = buf_count;
-    req_data->data = write_data;
-
-    r = uv_write(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, on_stream_write);
+    r = pyseq2uvbuf(seq, &views, &bufs, &buf_count);
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_StreamError);
-        goto error;
+        /* error is already set */
+        return NULL;
     }
 
-    Py_RETURN_NONE;
+    ret = pyuv_stream_write(self, views, bufs, buf_count, callback, NULL);
 
-error:
-    if (bufs) {
-        for (i = 0; i < buf_count; i++) {
-            PyMem_Free(bufs[i].base);
-        }
-        PyMem_Free(bufs);
-    }
-    if (write_data) {
-        PyMem_Free(write_data);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (wr) {
-        PyMem_Free(wr);
-    }
-    return NULL;
+    /* uv_write copies the uv_buf_t structures, so we can free them now */
+    PyMem_Free(bufs);
+
+    return ret;
 }
 
 
@@ -562,11 +380,10 @@ static PyObject *
 Stream_readable_get(Stream *self, void *closure)
 {
     UNUSED_ARG(closure);
-    if (!UV_HANDLE(self)) {
-        Py_RETURN_FALSE;
-    } else {
-        return PyBool_FromLong((long)uv_is_readable((uv_stream_t *)UV_HANDLE(self)));
-    }
+
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+
+    return PyBool_FromLong((long)uv_is_readable((uv_stream_t *)UV_HANDLE(self)));
 }
 
 
@@ -574,11 +391,10 @@ static PyObject *
 Stream_writable_get(Stream *self, void *closure)
 {
     UNUSED_ARG(closure);
-    if (!UV_HANDLE(self)) {
-        Py_RETURN_FALSE;
-    } else {
-        return PyBool_FromLong((long)uv_is_writable((uv_stream_t *)UV_HANDLE(self)));
-    }
+
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+
+    return PyBool_FromLong((long)uv_is_writable((uv_stream_t *)UV_HANDLE(self)));
 }
 
 
