@@ -1,7 +1,4 @@
 
-static PyObject* PyExc_TCPError;
-
-
 static void
 on_tcp_connection(uv_stream_t* server, int status)
 {
@@ -26,7 +23,7 @@ on_tcp_connection(uv_stream_t* server, int status)
 
     result = PyObject_CallFunctionObjArgs(self->on_new_connection_cb, self, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->on_new_connection_cb);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(py_errorno);
@@ -40,18 +37,14 @@ static void
 on_tcp_client_connection(uv_connect_t *req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    stream_req_data_t* req_data;
     TCP *self;
     PyObject *callback, *result, *py_errorno;
 
     ASSERT(req);
-    req_data = (stream_req_data_t *)req->data;
     self = (TCP *)req->handle->data;
-    callback = req_data->callback;
+    callback = (PyObject *)req->data;
 
     ASSERT(self);
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
 
     if (status != 0) {
         uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
@@ -63,16 +56,17 @@ on_tcp_client_connection(uv_connect_t *req, int status)
 
     result = PyObject_CallFunctionObjArgs(callback, self, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(callback);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(py_errorno);
 
     Py_DECREF(callback);
-    PyMem_Free(req_data);
     PyMem_Free(req);
 
+    /* Refcount was increased in the caller function */
     Py_DECREF(self);
+
     PyGILState_Release(gstate);
 }
 
@@ -82,9 +76,8 @@ TCP_func_bind(TCP *self, PyObject *args)
 {
     int r, bind_port, address_type;
     char *bind_ip;
-    struct in_addr addr4;
-    struct in6_addr addr6;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "(si):bind", &bind_ip, &bind_port)) {
@@ -96,11 +89,7 @@ TCP_func_bind(TCP *self, PyObject *args)
         return NULL;
     }
 
-    if (uv_inet_pton(AF_INET, bind_ip, &addr4) == 1) {
-        address_type = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, bind_ip, &addr6) == 1) {
-        address_type = AF_INET6;
-    } else {
+    if (pyuv_guess_ip_family(bind_ip, &address_type)) {
         PyErr_SetString(PyExc_ValueError, "invalid IP address");
         return NULL;
     }
@@ -129,6 +118,7 @@ TCP_func_listen(TCP *self, PyObject *args)
     backlog = 128;
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O|i:listen", &callback, &backlog)) {
@@ -166,6 +156,7 @@ TCP_func_accept(TCP *self, PyObject *args)
     int r;
     PyObject *client;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:accept", &client)) {
@@ -192,12 +183,10 @@ TCP_func_connect(TCP *self, PyObject *args)
 {
     int r, connect_port, address_type;
     char *connect_ip;
-    struct in_addr addr4;
-    struct in6_addr addr6;
     uv_connect_t *connect_req = NULL;
-    stream_req_data_t *req_data = NULL;
     PyObject *callback;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "(si)O:connect", &connect_ip, &connect_port, &callback)) {
@@ -214,30 +203,20 @@ TCP_func_connect(TCP *self, PyObject *args)
         return NULL;
     }
 
-    if (uv_inet_pton(AF_INET, connect_ip, &addr4) == 1) {
-        address_type = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, connect_ip, &addr6) == 1) {
-        address_type = AF_INET6;
-    } else {
+    if (pyuv_guess_ip_family(connect_ip, &address_type)) {
         PyErr_SetString(PyExc_ValueError, "invalid IP address");
         return NULL;
     }
 
-    connect_req = (uv_connect_t *)PyMem_Malloc(sizeof(uv_connect_t));
+    Py_INCREF(callback);
+
+    connect_req = PyMem_Malloc(sizeof *connect_req);
     if (!connect_req) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
-    req_data->callback = callback;
-    connect_req->data = (void *)req_data;
+    connect_req->data = (void *)callback;
 
     if (address_type == AF_INET) {
         r = uv_tcp_connect(connect_req, (uv_tcp_t *)UV_HANDLE(self), uv_ip4_addr(connect_ip, connect_port), on_tcp_client_connection);
@@ -250,16 +229,14 @@ TCP_func_connect(TCP *self, PyObject *args)
         goto error;
     }
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
 error:
-    if (connect_req) {
-        PyMem_Free(connect_req);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
+    Py_DECREF(callback);
+    PyMem_Free(connect_req);
     return NULL;
 }
 
@@ -275,6 +252,7 @@ TCP_func_getsockname(TCP *self)
 
     namelen = sizeof(sockname);
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_tcp_getsockname((uv_tcp_t *)UV_HANDLE(self), &sockname, &namelen);
@@ -309,6 +287,7 @@ TCP_func_getpeername(TCP *self)
 
     namelen = sizeof(peername);
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_tcp_getpeername((uv_tcp_t *)UV_HANDLE(self), &peername, &namelen);
@@ -340,6 +319,7 @@ TCP_func_nodelay(TCP *self, PyObject *args)
     int r;
     PyObject *enable;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:nodelay", &PyBool_Type, &enable)) {
@@ -363,6 +343,7 @@ TCP_func_keepalive(TCP *self, PyObject *args)
     unsigned int delay;
     PyObject *enable;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!I:keepalive", &PyBool_Type, &enable, &delay)) {
@@ -382,18 +363,10 @@ TCP_func_keepalive(TCP *self, PyObject *args)
 static PyObject *
 TCP_func_simultaneous_accepts(TCP *self, PyObject *args)
 {
-    /*
-    * This setting applies to Windows only.
-    * Enable/disable simultaneous asynchronous accept requests that are
-    * queued by the operating system when listening for new tcp connections.
-    * This setting is used to tune a tcp server for the desired performance.
-    * Having simultaneous accepts can significantly improve the rate of
-    * accepting connections (which is why it is enabled by default).
-    */
-
     int r;
     PyObject *enable;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:simultaneous_accepts", &PyBool_Type, &enable)) {
@@ -410,45 +383,45 @@ TCP_func_simultaneous_accepts(TCP *self, PyObject *args)
 }
 
 
+static PyObject *
+TCP_func_open(TCP *self, PyObject *args)
+{
+    long fd;
+
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    if (!PyArg_ParseTuple(args, "l:open", &fd)) {
+        return NULL;
+    }
+
+    uv_tcp_open((uv_tcp_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
+
+    Py_RETURN_NONE;
+}
+
+
 static int
 TCP_tp_init(TCP *self, PyObject *args, PyObject *kwargs)
 {
     int r;
-    uv_tcp_t *uv_tcp;
     Loop *loop;
-    PyObject *tmp = NULL;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_TCPError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!:__init__", &LoopType, &loop)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
-
-    uv_tcp = PyMem_Malloc(sizeof(uv_tcp_t));
-    if (!uv_tcp) {
-        PyErr_NoMemory();
-        Py_DECREF(loop);
-        return -1;
-    }
-
-    r = uv_tcp_init(UV_HANDLE_LOOP(self), (uv_tcp_t *)uv_tcp);
+    r = uv_tcp_init(loop->uv_loop, (uv_tcp_t *)UV_HANDLE(self));
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
-        Py_DECREF(loop);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_TCPError);
         return -1;
     }
-    uv_tcp->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_tcp;
+
+    initialize_handle(HANDLE(self), loop);
 
     return 0;
 }
@@ -457,10 +430,24 @@ TCP_tp_init(TCP *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 TCP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    TCP *self = (TCP *)StreamType.tp_new(type, args, kwargs);
-    if (!self) {
+    TCP *self;
+    uv_tcp_t *uv_tcp;
+
+    uv_tcp = PyMem_Malloc(sizeof *uv_tcp);
+    if (!uv_tcp) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (TCP *)StreamType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_tcp);
+        return NULL;
+    }
+
+    uv_tcp->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_tcp;
+
     return (PyObject *)self;
 }
 
@@ -493,6 +480,7 @@ TCP_tp_methods[] = {
     { "getpeername", (PyCFunction)TCP_func_getpeername, METH_NOARGS, "Get remote socket information." },
     { "nodelay", (PyCFunction)TCP_func_nodelay, METH_VARARGS, "Enable/disable Nagle's algorithm." },
     { "keepalive", (PyCFunction)TCP_func_keepalive, METH_VARARGS, "Enable/disable TCP keep-alive." },
+    { "open", (PyCFunction)TCP_func_open, METH_VARARGS, "Open the specified file descriptor and manage it as a TCP handle." },
     { "simultaneous_accepts", (PyCFunction)TCP_func_simultaneous_accepts, METH_VARARGS, "Enable/disable simultaneous asynchronous accept requests that are queued by the operating system when listening for new tcp connections." },
     { NULL }
 };
@@ -518,7 +506,7 @@ static PyTypeObject TCPType = {
     0,                                                             /*tp_getattro*/
     0,                                                             /*tp_setattro*/
     0,                                                             /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                       /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                                                             /*tp_doc*/
     (traverseproc)TCP_tp_traverse,                                 /*tp_traverse*/
     (inquiry)TCP_tp_clear,                                         /*tp_clear*/
