@@ -1,27 +1,17 @@
 
-static PyObject* PyExc_UDPError;
-
-
-typedef struct {
-    uv_buf_t *bufs;
-    int buf_count;
-    Py_buffer view;
-} udp_send_data_t;
-
 typedef struct {
     PyObject *callback;
-    udp_send_data_t *data;
-} udp_req_data_t;
+    Py_buffer *views;
+    int view_count;
+} udp_send_data_t;
 
 
 static uv_buf_t
 on_udp_alloc(uv_udp_t* handle, size_t suggested_size)
 {
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    uv_buf_t buf = uv_buf_init(PyMem_Malloc(suggested_size), suggested_size);
+    static char slab[PYUV_SLAB_SIZE];
     UNUSED_ARG(handle);
-    PyGILState_Release(gstate);
-    return buf;
+    return uv_buf_init(slab, sizeof(slab));
 }
 
 
@@ -63,7 +53,7 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
         data = PyBytes_FromStringAndSize(buf.base, nread);
         py_errorno = Py_None;
         Py_INCREF(Py_None);
-    } else if (nread < 0) {
+    } else {
         address_tuple = Py_None;
         Py_INCREF(Py_None);
         data = Py_None;
@@ -72,9 +62,9 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
         py_errorno = PyInt_FromLong((long)err.code);
     }
 
-    result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, address_tuple, data, py_errorno, NULL);
+    result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, address_tuple, PyInt_FromLong((long)flags), data, py_errorno, NULL);
     if (result == NULL) {
-        PyErr_WriteUnraisable(self->on_read_cb);
+        handle_uncaught_exception(HANDLE(self)->loop);
     }
     Py_XDECREF(result);
     Py_DECREF(address_tuple);
@@ -82,11 +72,6 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
     Py_DECREF(py_errorno);
 
 done:
-    /* In case of error libuv may not call alloc_cb */
-    if (buf.base != NULL) {
-        PyMem_Free(buf.base);
-    }
-
     Py_DECREF(self);
     PyGILState_Release(gstate);
 }
@@ -97,22 +82,18 @@ on_udp_send(uv_udp_send_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     int i;
-    udp_req_data_t* req_data;
-    udp_send_data_t* send_data;
+    udp_send_data_t* req_data;
     UDP *self;
     PyObject *callback, *result, *py_errorno;
 
     ASSERT(req);
 
-    req_data = (udp_req_data_t *)req->data;
-    send_data = req_data->data;
+    req_data = (udp_send_data_t *)req->data;
 
     self = (UDP *)req->handle->data;
     callback = req_data->callback;
 
     ASSERT(self);
-    /* Object could go out of scope in the callback, increase refcount to avoid it */
-    Py_INCREF(self);
 
     if (callback != Py_None) {
         if (status < 0) {
@@ -124,26 +105,23 @@ on_udp_send(uv_udp_send_t* req, int status)
         }
         result = PyObject_CallFunctionObjArgs(callback, self, py_errorno, NULL);
         if (result == NULL) {
-            PyErr_WriteUnraisable(callback);
+            handle_uncaught_exception(HANDLE(self)->loop);
         }
         Py_XDECREF(result);
         Py_DECREF(py_errorno);
     }
 
-    if (send_data->buf_count == 1) {
-        PyBuffer_Release(&send_data->view);
-    } else {
-        for (i = 0; i < send_data->buf_count; i++) {
-            PyMem_Free(send_data->bufs[i].base);
-        }
-        PyMem_Free(send_data->bufs);
-    }
-    PyMem_Free(send_data);
     Py_DECREF(callback);
+    for (i = 0; i < req_data->view_count; i++) {
+        PyBuffer_Release(&req_data->views[i]);
+    }
+    PyMem_Free(req_data->views);
     PyMem_Free(req_data);
     PyMem_Free(req);
 
+    /* Refcount was increased in the caller function */
     Py_DECREF(self);
+
     PyGILState_Release(gstate);
 }
 
@@ -151,14 +129,15 @@ on_udp_send(uv_udp_send_t* req, int status)
 static PyObject *
 UDP_func_bind(UDP *self, PyObject *args)
 {
-    int r, bind_port, address_type;
+    int r, bind_port, address_type, flags;
     char *bind_ip;
-    struct in_addr addr4;
-    struct in6_addr addr6;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "(si):bind", &bind_ip, &bind_port)) {
+    flags = 0;
+
+    if (!PyArg_ParseTuple(args, "(si)|i:bind", &bind_ip, &bind_port, &flags)) {
         return NULL;
     }
 
@@ -167,19 +146,15 @@ UDP_func_bind(UDP *self, PyObject *args)
         return NULL;
     }
 
-    if (uv_inet_pton(AF_INET, bind_ip, &addr4) == 1) {
-        address_type = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, bind_ip, &addr6) == 1) {
-        address_type = AF_INET6;
-    } else {
+    if (pyuv_guess_ip_family(bind_ip, &address_type)) {
         PyErr_SetString(PyExc_ValueError, "invalid IP address");
         return NULL;
     }
 
     if (address_type == AF_INET) {
-        r = uv_udp_bind((uv_udp_t *)UV_HANDLE(self), uv_ip4_addr(bind_ip, bind_port), 0);
+        r = uv_udp_bind((uv_udp_t *)UV_HANDLE(self), uv_ip4_addr(bind_ip, bind_port), flags);
     } else {
-        r = uv_udp_bind6((uv_udp_t *)UV_HANDLE(self), uv_ip6_addr(bind_ip, bind_port), UV_UDP_IPV6ONLY);
+        r = uv_udp_bind6((uv_udp_t *)UV_HANDLE(self), uv_ip6_addr(bind_ip, bind_port), flags);
     }
 
     if (r != 0) {
@@ -199,6 +174,7 @@ UDP_func_start_recv(UDP *self, PyObject *args)
 
     tmp = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start_recv", &callback)) {
@@ -230,6 +206,7 @@ UDP_func_stop_recv(UDP *self)
 {
     int r;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_udp_recv_stop((uv_udp_t *)UV_HANDLE(self));
@@ -248,109 +225,89 @@ UDP_func_stop_recv(UDP *self)
 static PyObject *
 UDP_func_send(UDP *self, PyObject *args)
 {
-    int r, buf_count, dest_port, address_type;
+    int r, dest_port, address_type;
     char *dest_ip;
-    struct in_addr addr4;
-    struct in6_addr addr6;
-    Py_buffer pbuf;
-    PyObject *callback;
-    uv_buf_t *bufs = NULL;
+    uv_buf_t buf;
+    Py_buffer *view;
+    PyObject *callback = Py_None;
     uv_udp_send_t *wr = NULL;
-    udp_req_data_t *req_data = NULL;
-    udp_send_data_t *write_data = NULL;
-
-    buf_count = 0;
-    callback = Py_None;
+    udp_send_data_t *req_data = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "(si)s*|O:send", &dest_ip, &dest_port, &pbuf, &callback)) {
+    view = PyMem_Malloc(sizeof *view);
+    if (!view) {
+        PyErr_NoMemory();
+        return NULL;
+    }
+
+#ifdef PYUV_PYTHON3
+    if (!PyArg_ParseTuple(args, "(si)y*|O:send", &dest_ip, &dest_port, view, &callback)) {
+#else
+    if (!PyArg_ParseTuple(args, "(si)s*|O:send", &dest_ip, &dest_port, view, &callback)) {
+#endif
         return NULL;
     }
 
     if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyBuffer_Release(&pbuf);
         PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
+        goto error1;
     }
 
     if (dest_port < 0 || dest_port > 65535) {
         PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
-        return NULL;
+        goto error1;
     }
 
-    if (uv_inet_pton(AF_INET, dest_ip, &addr4) == 1) {
-        address_type = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, dest_ip, &addr6) == 1) {
-        address_type = AF_INET6;
-    } else {
+    if (pyuv_guess_ip_family(dest_ip, &address_type)) {
         PyErr_SetString(PyExc_ValueError, "invalid IP address");
-        return NULL;
-    }
-
-    wr = (uv_udp_send_t *)PyMem_Malloc(sizeof(uv_udp_send_t));
-    if (!wr) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    req_data = (udp_req_data_t*) PyMem_Malloc(sizeof(udp_req_data_t));
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    write_data = (udp_send_data_t *) PyMem_Malloc(sizeof(udp_send_data_t));
-    if (!write_data) {
-        PyErr_NoMemory();
-        goto error;
+        goto error1;
     }
 
     Py_INCREF(callback);
-    req_data->callback = callback;
-    wr->data = (void *)req_data;
 
-    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t));
-    if (!bufs) {
+    wr = PyMem_Malloc(sizeof *wr);
+    if (!wr) {
         PyErr_NoMemory();
-        goto error;
+        goto error2;
     }
 
-    bufs[0] = uv_buf_init(pbuf.buf, pbuf.len);
-    buf_count = 1;
+    req_data = PyMem_Malloc(sizeof *req_data);
+    if (!req_data) {
+        PyErr_NoMemory();
+        goto error2;
+    }
 
-    write_data->bufs = bufs;
-    write_data->buf_count = buf_count;
-    write_data->view = pbuf;
-    req_data->data = write_data;
+    buf = uv_buf_init(view->buf, view->len);
+    req_data->callback = callback;
+    req_data->view_count = 1;
+    req_data->views = view;
+
+    wr->data = (void *)req_data;
 
     if (address_type == AF_INET) {
-        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), &buf, 1, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
     } else {
-        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), &buf, 1, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
     }
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
-        goto error;
+        goto error2;
     }
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
-error:
-    PyBuffer_Release(&pbuf);
-    if (bufs) {
-        PyMem_Free(bufs);
-    }
-    if (write_data) {
-        PyMem_Free(write_data);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (wr) {
-        PyMem_Free(wr);
-    }
+error2:
+    Py_DECREF(callback);
+    PyMem_Free(req_data);
+    PyMem_Free(wr);
+error1:
+    PyBuffer_Release(view);
+    PyMem_Free(view);
     return NULL;
 }
 
@@ -359,24 +316,16 @@ static PyObject *
 UDP_func_sendlines(UDP *self, PyObject *args)
 {
     int i, r, buf_count, dest_port, address_type;
-    char *data_str, *tmp, *dest_ip;
-    const char *default_encoding;
-    struct in_addr addr4;
-    struct in6_addr addr6;
-    Py_buffer pbuf;
-    Py_ssize_t data_len, n;
-    PyObject *callback, *seq, *iter, *item, *encoded;
-    uv_buf_t tmpbuf;
-    uv_buf_t *bufs, *new_bufs;
+    char *dest_ip;
+    PyObject *callback, *seq;
+    Py_buffer *views;
+    uv_buf_t *bufs;
     uv_udp_send_t *wr = NULL;
-    udp_req_data_t *req_data = NULL;
-    udp_send_data_t *write_data = NULL;
+    udp_send_data_t *req_data = NULL;
 
-    buf_count = 0;
-    bufs = new_bufs = NULL;
     callback = Py_None;
-    default_encoding = PyUnicode_GetDefaultEncoding();
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "(si)O|O:sendlines", &dest_ip, &dest_port, &seq, &callback)) {
@@ -393,165 +342,65 @@ UDP_func_sendlines(UDP *self, PyObject *args)
         return NULL;
     }
 
-    if (uv_inet_pton(AF_INET, dest_ip, &addr4) == 1) {
-        address_type = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, dest_ip, &addr6) == 1) {
-        address_type = AF_INET6;
-    } else {
+    if (pyuv_guess_ip_family(dest_ip, &address_type)) {
         PyErr_SetString(PyExc_ValueError, "invalid IP address");
         return NULL;
     }
 
-    wr = (uv_udp_send_t *)PyMem_Malloc(sizeof(uv_udp_send_t));
+    r = pyseq2uvbuf(seq, &views, &bufs, &buf_count);
+    if (r != 0) {
+        /* error is already set */
+        return NULL;
+    }
+
+    Py_INCREF(callback);
+
+    wr = PyMem_Malloc(sizeof *wr);
     if (!wr) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (udp_req_data_t*) PyMem_Malloc(sizeof(udp_req_data_t));
+    req_data = PyMem_Malloc(sizeof *req_data);
     if (!req_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    write_data = (udp_send_data_t *) PyMem_Malloc(sizeof(udp_send_data_t));
-    if (!write_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    Py_INCREF(callback);
     req_data->callback = callback;
+    req_data->view_count = buf_count;
+    req_data->views = views;
     wr->data = (void *)req_data;
 
-    iter = PyObject_GetIter(seq);
-    if (iter == NULL) {
-        goto error;
-    }
-
-    n = iter_guess_size(iter, 8);   /* if we can't get the size hint, preallocate 8 slots */
-    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t) * n);
-    if (!bufs) {
-        Py_DECREF(iter);
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    i = 0;
-    while (1) {
-        item = PyIter_Next(iter);
-        if (item == NULL) {
-            if (PyErr_Occurred()) {
-                Py_DECREF(iter);
-                goto error;
-            } else {
-                /* StopIteration */
-                break;
-            }
-        }
-
-        if (PyUnicode_Check(item)) {
-            encoded = PyUnicode_AsEncodedString(item, default_encoding, "strict");
-            if (encoded == NULL) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                goto error;
-            }
-            data_str = PyBytes_AS_STRING(encoded);
-            data_len = PyBytes_GET_SIZE(encoded);
-            tmp = (char *) PyMem_Malloc(data_len);
-            if (!tmp) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyErr_NoMemory();
-                goto error;
-            }
-            memcpy(tmp, data_str, data_len);
-            tmpbuf = uv_buf_init(tmp, data_len);
-        } else {
-            if (PyObject_GetBuffer(item, &pbuf, PyBUF_CONTIG_RO) < 0) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                goto error;
-            }
-            data_str = pbuf.buf;
-            data_len = pbuf.len;
-            tmp = (char *) PyMem_Malloc(data_len);
-            if (!tmp) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyBuffer_Release(&pbuf);
-                PyErr_NoMemory();
-                goto error;
-            }
-            memcpy(tmp, data_str, data_len);
-            tmpbuf = uv_buf_init(tmp, data_len);
-            PyBuffer_Release(&pbuf);
-        }
-
-        /* Check if we allocated enough space */
-        if (buf_count+1 < n) {
-            /* we have enough size */
-        } else {
-            /* preallocate 8 more slots */
-            n += 8;
-            new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * n);
-            if (!new_bufs) {
-                Py_DECREF(item);
-                Py_DECREF(iter);
-                PyErr_NoMemory();
-                goto error;
-            }
-            bufs = new_bufs;
-        }
-        bufs[i] = tmpbuf;
-        i++;
-        buf_count++;
-        Py_DECREF(item);
-    }
-    Py_DECREF(iter);
-
-    /* we may have over allocated space, shrink it to the minimum required */
-    new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * buf_count);
-    if (!new_bufs) {
-        PyErr_NoMemory();
-        goto error;
-    }
-    bufs = new_bufs;
-
-    write_data->bufs = bufs;
-    write_data->buf_count = buf_count;
-    req_data->data = write_data;
-
     if (address_type == AF_INET) {
         r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
     } else {
         r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
     }
+
+    /* uv_write copies the uv_buf_t structures, so we can free them now */
+    PyMem_Free(bufs);
+    bufs = NULL;
+
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         goto error;
     }
 
+    /* Increase refcount so that object is not removed before the callback is called */
+    Py_INCREF(self);
+
     Py_RETURN_NONE;
 
 error:
-    if (bufs) {
-        for (i = 0; i < buf_count; i++) {
-            PyMem_Free(bufs[i].base);
-        }
-        PyMem_Free(bufs);
-    }
-    if (write_data) {
-        PyMem_Free(write_data);
-    }
-    if (req_data) {
-        Py_DECREF(callback);
-        PyMem_Free(req_data);
-    }
-    if (wr) {
-        PyMem_Free(wr);
+    Py_DECREF(callback);
+    for (i = 0; i < buf_count; i++) {
+        PyBuffer_Release(&views[i]);
     }
+    PyMem_Free(views);
+    PyMem_Free(bufs);
+    PyMem_Free(req_data);
+    PyMem_Free(wr);
     return NULL;
 }
 
@@ -564,6 +413,7 @@ UDP_func_set_membership(UDP *self, PyObject *args)
 
     interface_address = NULL;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "si|s:set_membership", &multicast_address, &membership, &interface_address)) {
@@ -591,6 +441,7 @@ UDP_func_getsockname(UDP *self)
 
     namelen = sizeof(sockname);
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     r = uv_udp_getsockname((uv_udp_t *)UV_HANDLE(self), &sockname, &namelen);
@@ -619,6 +470,7 @@ UDP_func_set_multicast_ttl(UDP *self, PyObject *args)
 {
     int r, ttl;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:set_multicast_ttl", &ttl)) {
@@ -646,6 +498,7 @@ UDP_func_set_broadcast(UDP *self, PyObject *args)
     int r;
     PyObject *enable;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:set_broadcast", &PyBool_Type, &enable)) {
@@ -668,6 +521,7 @@ UDP_func_set_multicast_loop(UDP *self, PyObject *args)
     int r;
     PyObject *enable;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:set_multicast_loop", &PyBool_Type, &enable)) {
@@ -689,6 +543,7 @@ UDP_func_set_ttl(UDP *self, PyObject *args)
 {
     int r, ttl;
 
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:set_multicast_ttl", &ttl)) {
@@ -710,44 +565,45 @@ UDP_func_set_ttl(UDP *self, PyObject *args)
 }
 
 
+static PyObject *
+UDP_func_open(UDP *self, PyObject *args)
+{
+    long fd;
+
+    RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    if (!PyArg_ParseTuple(args, "l:open", &fd)) {
+        return NULL;
+    }
+
+    uv_udp_open((uv_udp_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
+
+    Py_RETURN_NONE;
+}
+
+
 static int
 UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
 {
     int r;
-    uv_udp_t *uv_udp_handle = NULL;
     Loop *loop;
-    PyObject *tmp = NULL;
 
     UNUSED_ARG(kwargs);
 
-    if (UV_HANDLE(self)) {
-        PyErr_SetString(PyExc_UDPError, "Object already initialized");
-        return -1;
-    }
+    RAISE_IF_HANDLE_INITIALIZED(self, -1);
 
     if (!PyArg_ParseTuple(args, "O!:__init__", &LoopType, &loop)) {
         return -1;
     }
 
-    tmp = (PyObject *)((Handle *)self)->loop;
-    Py_INCREF(loop);
-    ((Handle *)self)->loop = loop;
-    Py_XDECREF(tmp);
-
-    uv_udp_handle = PyMem_Malloc(sizeof(uv_udp_t));
-    if (!uv_udp_handle) {
-        PyErr_NoMemory();
-        Py_DECREF(loop);
-        return -1;
-    }
-    r = uv_udp_init(UV_HANDLE_LOOP(self), uv_udp_handle);
+    r = uv_udp_init(loop->uv_loop, (uv_udp_t *)UV_HANDLE(self));
     if (r != 0) {
-        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
-        Py_DECREF(loop);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_UDPError);
         return -1;
     }
-    uv_udp_handle->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_udp_handle;
+
+    initialize_handle(HANDLE(self), loop);
 
     return 0;
 }
@@ -756,10 +612,24 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 UDP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    UDP *self = (UDP *)HandleType.tp_new(type, args, kwargs);
-    if (!self) {
+    UDP *self;
+    uv_udp_t *uv_udp;
+
+    uv_udp = PyMem_Malloc(sizeof *uv_udp);
+    if (!uv_udp) {
+        PyErr_NoMemory();
         return NULL;
     }
+
+    self = (UDP *)HandleType.tp_new(type, args, kwargs);
+    if (!self) {
+        PyMem_Free(uv_udp);
+        return NULL;
+    }
+
+    uv_udp->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_udp;
+
     return (PyObject *)self;
 }
 
@@ -790,6 +660,7 @@ UDP_tp_methods[] = {
     { "send", (PyCFunction)UDP_func_send, METH_VARARGS, "Send data over UDP." },
     { "sendlines", (PyCFunction)UDP_func_sendlines, METH_VARARGS, "Send a sequence of data over UDP." },
     { "getsockname", (PyCFunction)UDP_func_getsockname, METH_NOARGS, "Get local socket information." },
+    { "open", (PyCFunction)UDP_func_open, METH_VARARGS, "Open the specified file descriptor and manage it as a UDP handle." },
     { "set_membership", (PyCFunction)UDP_func_set_membership, METH_VARARGS, "Set membership for multicast address." },
     { "set_multicast_ttl", (PyCFunction)UDP_func_set_multicast_ttl, METH_VARARGS, "Set the multicast TTL." },
     { "set_multicast_loop", (PyCFunction)UDP_func_set_multicast_loop, METH_VARARGS, "Set IP multicast loop flag. Makes multicast packets loop back to local sockets." },
@@ -819,7 +690,7 @@ static PyTypeObject UDPType = {
     0,                                                              /*tp_getattro*/
     0,                                                              /*tp_setattro*/
     0,                                                              /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,  /*tp_flags*/
     0,                                                              /*tp_doc*/
     (traverseproc)UDP_tp_traverse,                                  /*tp_traverse*/
     (inquiry)UDP_tp_clear,                                          /*tp_clear*/
@@ -840,4 +711,3 @@ static PyTypeObject UDPType = {
     UDP_tp_new,                                                     /*tp_new*/
 };
 
-
