@@ -37,11 +37,8 @@ class TimerTest(unittest2.TestCase):
             timer.stop()
             timer.close()
         loop = pyuv.Loop.default_loop()
-        timer = pyuv.Timer(loop)
-        loop.unref()
+        self.timer = pyuv.Timer(loop)
         loop.run()
-        # When timer is destroyed it will unref the loop again, so loop.run() would block next time
-        loop.ref()
         self.assertEqual(self.timer_cb_called, 0)
 
     def test_timer_ref2(self):
@@ -51,13 +48,14 @@ class TimerTest(unittest2.TestCase):
             timer.stop()
             timer.close()
         loop = pyuv.Loop.default_loop()
-        timer = pyuv.Timer(loop)
-        timer.start(timer_cb, 0.1, 0)
-        loop.unref()
+        self.timer = pyuv.Timer(loop)
+        self.timer.start(timer_cb, 0.1, 0)
+        self.timer.unref()
         loop.run()
-        # When timer is destroyed it will unref the loop again, so loop.run() would block next time
-        loop.ref()
         self.assertEqual(self.timer_cb_called, 0)
+        self.timer.ref()
+        loop.run()
+        self.assertEqual(self.timer_cb_called, 1)
 
 
 if __name__ == '__main__':
